package checkers.types;

import checkers.quals.Default;
import checkers.nullness.quals.Nullable;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.*;

/**
 * Creates new annotations as {@link AnnotationMirror}s that may be added to
 * {@link AnnotatedTypeMirror}s.
 */
@Default("checkers.nullness.quals.NonNull")
public class AnnotationFactory {

    /** Element utilities for acquiring {@link TypeMirror}s from strings. */
    private final Elements elements;

    /**
     * Creates an annotation factory.
     *
     * @param env the {@link ProcessingEnvironment} for this factory
     */
    public AnnotationFactory(ProcessingEnvironment env) {
        this.elements = env.getElementUtils();
    }

    /** Caching for annotation creation. */
    private static final Map<CharSequence, AnnotationMirror> annotationsFromNames
        = new HashMap<CharSequence, AnnotationMirror>();

    /**
     * Creates an {@link AnnotationMirror} given by a particular
     * fully-qualified name.
     *
     * @param name the name of the annotation to create
     * @return an {@link AnnotationMirror} of type {@code} name
     */
    public AnnotationMirror fromName(CharSequence name) {
        if (annotationsFromNames.containsKey(name))
            return annotationsFromNames.get(name);
        final DeclaredType annoType = typeFromName(name);
        if (annoType == null
                || annoType.asElement().getKind() != ElementKind.ANNOTATION_TYPE)
            throw new AssertionError(annoType + " is not an annotation");
        AnnotationMirror result = new AnnotationMirror() {
            @Override
            public DeclaredType getAnnotationType() {
                return annoType;
            }
            @Override
            public Map<? extends ExecutableElement, ? extends AnnotationValue>
                getElementValues() {
                return Collections.emptyMap();
            }
            @Override
            public String toString() {
                return "@" + annoType;
            }
        };
        annotationsFromNames.put(name, result);
        return result;
    }

    /**
     * Creates an {@link AnnotationMirror} given by a particular annotation
     * class.
     *
     * @param clazz the annotation class
     * @return an {@link AnnotationMirror} of type given type
     */
    public AnnotationMirror fromClass(Class<? extends Annotation> clazz) {
        return fromName(clazz.getCanonicalName());
    }

    /**
     * A utility method that converts a {@link CharSequence} (usually a {@link
     * String}) into a {@link TypeMirror} named thereby.
     *
     * @param name the name of a type
     * @return the {@link TypeMirror} corresponding to that name
     */
    private DeclaredType typeFromName(CharSequence name) {

        /*@Nullable*/ TypeElement typeElt = elements.getTypeElement(name);
        if (typeElt == null)
            throw new IllegalArgumentException("invalid name: " + name);

        return (DeclaredType)typeElt.asType();
    }
}
